from .libindy import do_call, create_cb

from ctypes import *

import logging


async def create_and_store_my_did(wallet_handle: int,
                                  did_json: str) -> (str, str, str):
    """
    Creates keys (signing and encryption keys) for a new
    DID (owned by the caller of the library).
    Identity's DID must be either explicitly provided, or taken as the first 16 bit of verkey.
    Saves the Identity DID with keys in a secured Wallet, so that it can be used to sign
    and encrypt transactions.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did_json: Identity information as json. Example:
        {
            "did": string, (optional;
                    if not provided and cid param is false then the first 16 bit of the verkey will be used as a new DID;
                    if not provided and cid is true then the full verkey will be used as a new DID;
                    if provided, then keys will be replaced - key rotation use case)
            "seed": string, (optional; if not provide then a random one will be created)
            "crypto_type": string, (optional; if not set then ed25519 curve is used;
                      currently only 'ed25519' value is supported for this field)
            "cid": bool, (optional; if not set then false is used;)
        }
    :return: DID, verkey (for verification of signature) and public_key (for decryption)
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_and_store_my_did: >>> wallet_handle: %r, did_json: %r",
                 wallet_handle,
                 did_json)

    if not hasattr(create_and_store_my_did, "cb"):
        logger.debug("create_wallet: Creating callback")
        create_and_store_my_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did_json = c_char_p(did_json.encode('utf-8'))

    did, verkey, pk = await do_call('indy_create_and_store_my_did',
                                    c_wallet_handle,
                                    c_did_json,
                                    create_and_store_my_did.cb)

    res = (did.decode(), verkey.decode(), pk.decode())

    logger.debug("create_and_store_my_did: <<< res: %r", res)
    return res


async def replace_keys(wallet_handle: int,
                       did: str,
                       identity_json: str) -> (str, str):
    """
    Generated new keys (signing and encryption keys) for an existing
    DID (owned by the caller of the library).

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: signing DID
    :param identity_json: Identity information as json. Example:
        {
            "seed": string, (optional; if not provide then a random one will be created)
            "crypto_type": string, (optional; if not set then ed25519 curve is used;
                      currently only 'ed25519' value is supported for this field)
        }
    :return: verkey (for verification of signature) and public_key (for decryption)
    """

    logger = logging.getLogger(__name__)
    logger.debug("replace_keys: >>> wallet_handle: %r, did: %r, identity_json: %r",
                 wallet_handle,
                 did,
                 identity_json)

    if not hasattr(replace_keys, "cb"):
        logger.debug("replace_keys: Creating callback")
        replace_keys.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_identity_json = c_char_p(identity_json.encode('utf-8'))

    verkey, pk = await do_call('indy_replace_keys',
                               c_wallet_handle,
                               c_did,
                               c_identity_json,
                               replace_keys.cb)

    res = (verkey.decode(), pk.decode())

    logger.debug("replace_keys: <<< res: %r", res)
    return res


async def store_their_did(wallet_handle: int,
                          identity_json: str) -> None:
    """
    Saves their DID for a pairwise connection in a secured Wallet,
    so that it can be used to verify transaction.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param identity_json: Identity information as json. Example:
        {
           "did": string, (required)
           "verkey": string (optional, if only pk is provided),
           "crypto_type": string, (optional; if not set then ed25519 curve is used;
                  currently only 'ed25519' value is supported for this field)
        }
    :return: None
    """

    logger = logging.getLogger(__name__)
    logger.debug("store_their_did: >>> wallet_handle: %r, identity_json: %r",
                 wallet_handle,
                 identity_json)

    if not hasattr(store_their_did, "cb"):
        logger.debug("store_their_did: Creating callback")
        store_their_did.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_identity_json = c_char_p(identity_json.encode('utf-8'))

    res = await do_call('indy_store_their_did',
                        c_wallet_handle,
                        c_identity_json,
                        store_their_did.cb)

    logger.debug("store_their_did: <<< res: %r", res)
    return res


async def sign(wallet_handle: int,
               did: str,
               msg: bytes) -> bytes:
    """
    Signs a message by a signing key associated with my DID. The DID with a signing key
    must be already created and stored in a secured wallet (see create_and_store_my_identity)

    :param wallet_handle: wallet handler (created by open_wallet).
    :param did: signing DID
    :param msg: a message to be signed
    :return: a signature string
    """

    logger = logging.getLogger(__name__)
    logger.debug("sign: >>> wallet_handle: %r, did: %r, msg: %r",
                 wallet_handle,
                 did,
                 msg)

    def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32):
        return bytes(arr_ptr[:arr_len]),

    if not hasattr(sign, "cb"):
        logger.debug("sign: Creating callback")
        sign.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, POINTER(c_uint8), c_uint32), transform_cb)

    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_msg_len = c_uint32(len(msg))

    signature = await do_call('indy_sign',
                              c_wallet_handle,
                              c_did,
                              msg,
                              c_msg_len,
                              sign.cb)

    logger.debug("sign: <<< res: %r", signature)
    return signature


async def verify_signature(wallet_handle: int,
                           pool_handle: int,
                           did: str,
                           msg: bytes,
                           signature: bytes) -> bool:
    """
    Verify a signature created by a key associated with a DID.
    If a secure wallet doesn't contain a verkey associated with the given DID,
    then verkey is read from the Ledger.
    Otherwise either an existing verkey from wallet is used (see wallet_store_their_identity),
    or it checks the Ledger (according to freshness settings set during initialization)
    whether verkey is still the same and updates verkey for the DID if needed.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param pool_handle: pool handle.
    :param did: DID that signed the message
    :param msg: message
    :param signature: a signature to be verified
    :return: valid: true - if signature is valid, false - otherwise
    """

    logger = logging.getLogger(__name__)
    logger.debug("verify_signature: >>> wallet_handle: %r, pool_handle: %r, did: %r, signed_msg: %r",
                 wallet_handle,
                 pool_handle,
                 did,
                 msg,
                 signature)

    if not hasattr(verify_signature, "cb"):
        logger.debug("verify_signature: Creating callback")
        verify_signature.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_bool))

    c_wallet_handle = c_int32(wallet_handle)
    c_pool_handle = c_int32(pool_handle)
    c_did = c_char_p(did.encode('utf-8'))
    c_msg_len = c_uint32(len(msg))
    c_signature_len = c_uint32(len(signature))

    res = await do_call('indy_verify_signature',
                        c_wallet_handle,
                        c_pool_handle,
                        c_did,
                        msg,
                        c_msg_len,
                        signature,
                        c_signature_len,
                        verify_signature.cb)

    logger.debug("verify_signature: <<< res: %r", res)
    return res


async def encrypt(wallet_handle: int,
                  pool_handle: int,
                  my_did: str,
                  did: str,
                  msg: bytes) -> (bytes, bytes):
    """
    Encrypts a message by a public key associated with a DID.
    If a secure wallet doesn't contain a public key associated with the given DID,
    then the public key is read from the Ledger.
    Otherwise either an existing public key from wallet is used (see wallet_store_their_identity),
    or it checks the Ledger (according to freshness settings set during initialization)
    whether public key is still the same and updates public key for the DID if needed.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param pool_handle: pool handle.
    :param my_did: encrypting DID
    :param did: encrypting DID
    :param msg: a message to be signed
    :return: an encrypted message and nonce
    """

    logger = logging.getLogger(__name__)
    logger.debug("encrypt: >>> wallet_handle: %r, pool_handle: %r, my_did: %r, did: %r, msg: %r",
                 wallet_handle,
                 pool_handle,
                 my_did,
                 did,
                 msg)

    def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32, arr_ptr2: POINTER(c_uint8), arr_len2: c_uint32):
        return bytes(arr_ptr[:arr_len]), bytes(arr_ptr2[:arr_len2])

    if not hasattr(encrypt, "cb"):
        logger.debug("encrypt: Creating callback")
        encrypt.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, POINTER(c_uint8), c_uint32, POINTER(c_uint8),
                                         c_uint32), transform_cb)

    c_wallet_handle = c_int32(wallet_handle)
    c_pool_handle = c_int32(pool_handle)
    c_my_did = c_char_p(my_did.encode('utf-8'))
    c_did = c_char_p(did.encode('utf-8'))
    c_msg_len = c_uint32(len(msg))

    encrypted_message, nonce = await do_call('indy_encrypt',
                                             c_wallet_handle,
                                             c_pool_handle,
                                             c_my_did,
                                             c_did,
                                             msg,
                                             c_msg_len,
                                             encrypt.cb)
    res = (encrypted_message, nonce)

    logger.debug("encrypt: <<< res: %r", res)
    return res


async def decrypt(wallet_handle: int,
                  my_did: str,
                  did: str,
                  encrypted_msg: bytes,
                  nonce: bytes) -> bytes:
    """
    Decrypts a message encrypted by a public key associated with my DID.
    The DID with a secret key must be already created and
    stored in a secured wallet (see wallet_create_and_store_my_identity)

    :param wallet_handle: wallet handler (created by open_wallet).
    :param my_did: DID
    :param did: DID that signed the message
    :param encrypted_msg: encrypted message
    :param nonce: nonce that encrypted message
    :return: decrypted message
    """

    logger = logging.getLogger(__name__)
    logger.debug("decrypt: >>> wallet_handle: %r, my_did: %r, did: %r, encrypted_msg: %r, nonce: %r",
                 wallet_handle,
                 my_did,
                 did,
                 encrypted_msg,
                 nonce)

    def transform_cb(arr_ptr: POINTER(c_uint8), arr_len: c_uint32):
        return bytes(arr_ptr[:arr_len]),

    if not hasattr(decrypt, "cb"):
        logger.debug("decrypt: Creating callback")
        decrypt.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, POINTER(c_uint8), c_uint32), transform_cb)

    c_wallet_handle = c_int32(wallet_handle)
    c_my_did = c_char_p(my_did.encode('utf-8'))
    c_did = c_char_p(did.encode('utf-8'))
    c_encrypted_msg_len = c_uint32(len(encrypted_msg))
    c_nonce_len = c_uint32(len(nonce))

    decrypted_message = await do_call('indy_decrypt',
                                      c_wallet_handle,
                                      c_my_did,
                                      c_did,
                                      encrypted_msg,
                                      c_encrypted_msg_len,
                                      nonce,
                                      c_nonce_len,
                                      decrypt.cb)

    logger.debug("decrypt: <<< res: %r", decrypted_message)
    return decrypted_message



from .libindy import do_call, create_cb

from typing import Optional
from ctypes import *

import logging


async def create_wallet(pool_name: str,
                        name: str,
                        xtype: Optional[str],
                        config: Optional[str],
                        credentials: Optional[str]) -> None:
    """
    Creates a new secure wallet with the given unique name.

    :param pool_name: Name of the pool that corresponds to this wallet.
    :param name: Name of the wallet.
    :param xtype: (optional) Type of the wallet. Defaults to 'default'.
     Custom types can be registered with indy_register_wallet_type call.
    :param config: (optional) Wallet configuration json. List of supported keys are defined by wallet type.
     if NULL, then default config will be used.
    :param credentials: (optional) Wallet credentials json. List of supported keys are defined by wallet type.
     if NULL, then default config will be used.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("create_wallet: >>> pool_name: %r, name: %r, xtype: %r, config: %r, credentials: %r",
                 pool_name,
                 name,
                 xtype,
                 config,
                 credentials)

    if not hasattr(create_wallet, "cb"):
        logger.debug("create_wallet: Creating callback")
        create_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_pool_name = c_char_p(pool_name.encode('utf-8'))
    c_name = c_char_p(name.encode('utf-8'))
    c_xtype = c_char_p(xtype.encode('utf-8')) if xtype is not None else None
    c_config = c_char_p(config.encode('utf-8')) if config is not None else None
    c_credentials = c_char_p(credentials.encode('utf-8')) if credentials is not None else None

    await do_call('indy_create_wallet',
                  c_pool_name,
                  c_name,
                  c_xtype,
                  c_config,
                  c_credentials,
                  create_wallet.cb)

    logger.debug("create_wallet: <<<")


async def open_wallet(name: str,
                      runtime_config: Optional[str],
                      credentials: Optional[str]) -> int:
    """
    Opens the wallet with specific name.
    Wallet with corresponded name must be previously created with indy_create_wallet method.
    It is impossible to open wallet with the same name more than once.

    :param name: Name of the wallet.
    :param runtime_config: (optional) Runtime wallet configuration json.
     if NULL, then default runtime_config will be used. Example:
        {
            "freshness_time": string (optional), Amount of minutes to consider wallet value as fresh. Defaults to 24*60.
            ... List of additional supported keys are defined by wallet type.
        }
    :param credentials: (optional) Wallet credentials json. List of supported keys are defined by wallet type.
     if NULL, then default credentials will be used.
    :return: Handle to opened wallet to use in methods that require wallet access.
    """

    logger = logging.getLogger(__name__)
    logger.debug("open_wallet: >>> name: %r, runtime_config: %r, credentials: %r",
                 name,
                 runtime_config,
                 credentials)

    if not hasattr(open_wallet, "cb"):
        logger.debug("open_wallet: Creating callback")
        open_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_int32))

    c_name = c_char_p(name.encode('utf-8'))
    c_runtime_config = c_char_p(runtime_config.encode('utf-8')) if runtime_config is not None else None
    c_credentials = c_char_p(credentials.encode('utf-8')) if credentials is not None else None

    res = await do_call('indy_open_wallet',
                        c_name,
                        c_runtime_config,
                        c_credentials,
                        open_wallet.cb)

    logger.debug("open_wallet: <<< res: %r", res)
    return res


async def close_wallet(handle: int) -> None:
    """
    Closes opened wallet and frees allocated resources.

    :param handle: wallet handle returned by indy_open_wallet.
    :return: Error code
    """

    logger = logging.getLogger(__name__)
    logger.debug("close_wallet: >>> handle: %i", handle)

    if not hasattr(close_wallet, "cb"):
        logger.debug("close_wallet: Creating callback")
        close_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_handle = c_int32(handle)

    await do_call('indy_close_wallet',
                  c_handle,
                  close_wallet.cb)

    logger.debug("close_wallet: <<<")


# pub extern fn indy_delete_wallet(command_handle: i32,
#                                    name: *const c_char,
#                                    credentials: *const c_char,
#                                    cb: Option<extern fn(xcommand_handle: i32, err: ErrorCode)>) -> ErrorCode {
async def delete_wallet(name: str,
                        credentials: Optional[str]) -> None:
    """
    Deletes created wallet.

    :param name: Name of the wallet to delete.
    :param credentials: (optional) Wallet credentials json. List of supported keys are defined by wallet type.
     if NULL, then default credentials will be used.
    :return:
    """

    logger = logging.getLogger(__name__)
    logger.debug("delete_wallet: >>> name: %r, credentials: %r",
                 name,
                 credentials)

    if not hasattr(delete_wallet, "cb"):
        logger.debug("delete_wallet: Creating callback")
        delete_wallet.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_name = c_char_p(name.encode('utf-8'))
    c_credentials = c_char_p(credentials.encode('utf-8')) if credentials is not None else None

    await do_call('indy_delete_wallet',
                  c_name,
                  c_credentials,
                  delete_wallet.cb)

    logger.debug("delete_wallet: <<<")



from .error import ErrorCode, IndyError

from ctypes import *

import asyncio
import sys
import itertools
import logging

_futures = {}
_futures_counter = itertools.count()


def do_call(name: str, *args):
    logger = logging.getLogger(__name__)
    logger.debug("do_call: >>> name: %s, args: %s", name, args)

    event_loop = asyncio.get_event_loop()
    future = event_loop.create_future()
    command_handle = next(_futures_counter)

    _futures[command_handle] = (event_loop, future)

    err = getattr(_cdll(), name)(command_handle,
                                 *args)

    logger.debug("do_call: Function %s returned err: %i", name, err)

    if err != ErrorCode.Success:
        logger.warning("_do_call: Function %s returned error %i", name, err)
        future.set_exception(IndyError(ErrorCode(err)))

    logger.debug("do_call: <<< %s", future)
    return future


def create_cb(cb_type: CFUNCTYPE, transform_fn=None):
    logger = logging.getLogger(__name__)
    logger.debug("create_cb: >>> cb_type: %s", cb_type)

    def _cb(command_handle: int, err: int, *args):
        if transform_fn:
            args = transform_fn(*args)
        _indy_callback(command_handle, err, *args)

    res = cb_type(_cb)

    logger.debug("create_cb: <<< res: %s", res)
    return res


def _indy_callback(command_handle: int, err: int, *args):
    logger = logging.getLogger(__name__)
    logger.debug("_indy_callback: >>> command_handle: %i, err %i, args: %s", command_handle, err, args)

    (event_loop, future) = _futures[command_handle]
    event_loop.call_soon_threadsafe(_indy_loop_callback, command_handle, err, *args)

    logger.debug("_indy_callback: <<<")


def _indy_loop_callback(command_handle: int, err, *args):
    logger = logging.getLogger(__name__)
    logger.debug("_indy_loop_callback: >>> command_handle: %i, err %i, args: %s", command_handle, err, args)

    (event_loop, future) = _futures.pop(command_handle)

    if err != ErrorCode.Success:
        logger.warning("_indy_loop_callback: Function returned error %i", err)
        future.set_exception(IndyError(ErrorCode(err)))
    else:
        if len(args) == 0:
            res = None
        elif len(args) == 1:
            (res,) = args
        else:
            res = args

        logger.warning("_indy_loop_callback: Function returned %s", res)
        future.set_result(res)

    logger.debug("_indy_loop_callback <<<")


def _cdll() -> CDLL:
    if not hasattr(_cdll, "cdll"):
        _cdll.cdll = _load_cdll()

    return _cdll.cdll


def _load_cdll() -> CDLL:
    logger = logging.getLogger(__name__)
    logger.debug("_load_cdll: >>>")

    libindy_prefix_mapping = {"darwin": "lib", "linux": "lib", "linux2": "lib", "win32": ""}
    libindy_suffix_mapping = {"darwin": ".dylib", "linux": ".so", "linux2": ".so", "win32": ".dll"}

    os_name = sys.platform
    logger.debug("_load_cdll: Detected OS name: %s", os_name)

    try:
        libindy_prefix = libindy_prefix_mapping[os_name]
        libindy_suffix = libindy_suffix_mapping[os_name]
    except KeyError:
        logger.error("_load_cdll: OS isn't supported: %s", os_name)
        raise OSError("OS isn't supported: %s", os_name)

    libindy_name = "{0}indy{1}".format(libindy_prefix, libindy_suffix)
    logger.debug("_load_cdll: Resolved libindy name is: %s", libindy_name)

    try:
        res = CDLL(libindy_name)
        logger.debug("_load_cdll: <<< res: %s", res)
        return res
    except OSError as e:
        logger.error("_load_cdll: Can't load libindy: %s", e)
        raise e



from enum import IntEnum


class ErrorCode(IntEnum):
    Success = 0,

    # Common errors

    # Caller passed invalid value as param 1 (null, invalid json and etc..)
    CommonInvalidParam1 = 100,

    # Caller passed invalid value as param 2 (null, invalid json and etc..)
    CommonInvalidParam2 = 101,

    # Caller passed invalid value as param 3 (null, invalid json and etc..)
    CommonInvalidParam3 = 102,

    # Caller passed invalid value as param 4 (null, invalid json and etc..)
    CommonInvalidParam4 = 103,

    # Caller passed invalid value as param 5 (null, invalid json and etc..)
    CommonInvalidParam5 = 104,

    # Caller passed invalid value as param 6 (null, invalid json and etc..)
    CommonInvalidParam6 = 105,

    # Caller passed invalid value as param 7 (null, invalid json and etc..)
    CommonInvalidParam7 = 106,

    # Caller passed invalid value as param 8 (null, invalid json and etc..)
    CommonInvalidParam8 = 107,

    # Caller passed invalid value as param 9 (null, invalid json and etc..)
    CommonInvalidParam9 = 108,

    # Caller passed invalid value as param 10 (null, invalid json and etc..)
    CommonInvalidParam10 = 109,

    # Caller passed invalid value as param 11 (null, invalid json and etc..)
    CommonInvalidParam11 = 110,

    # Caller passed invalid value as param 12 (null, invalid json and etc..)
    CommonInvalidParam12 = 111,

    # Invalid library state was detected in runtime. It signals library bug
    CommonInvalidState = 112,

    # Object (json, config, key, claim and etc...) passed by library caller has invalid structure
    CommonInvalidStructure = 113,

    # IO Error
    CommonIOError = 114,

    # Wallet errors
    # Caller passed invalid wallet handle
    WalletInvalidHandle = 200,

    # Unknown type of wallet was passed on create_wallet
    WalletUnknownTypeError = 201,

    # Attempt to register already existing wallet type
    WalletTypeAlreadyRegisteredError = 202,

    # Attempt to create wallet with name used for another exists wallet
    WalletAlreadyExistsError = 203,

    # Requested entity id isn't present in wallet
    WalletNotFoundError = 204,

    # Trying to use wallet with pool that has different name
    WalletIncompatiblePoolError = 205,

    # Trying to open wallet that was opened already
    WalletAlreadyOpenedError = 206,

    # Ledger errors
    # Trying to open pool ledger that wasn't created before
    PoolLedgerNotCreatedError = 300,

    # Caller passed invalid pool ledger handle
    PoolLedgerInvalidPoolHandle = 301,

    # Pool ledger terminated
    PoolLedgerTerminated = 302,

    # No concensus during ledger operation
    LedgerNoConsensusError = 303,

    # Attempt to send unknown or incomplete transaction message
    LedgerInvalidTransaction = 304,

    # Attempt to send transaction without the necessary privileges
    LedgerSecurityError = 305,

    # Attempt to create pool ledger config with name used for another existing pool
    PoolLedgerConfigAlreadyExistsError = 306,

    # Revocation registry is full and creation of new registry is necessary
    AnoncredsRevocationRegistryFullError = 400,

    AnoncredsInvalidUserRevocIndex = 401,

    AnoncredsAccumulatorIsFull = 402,

    AnoncredsNotIssuedError = 403,

    # Attempt to generate master secret with dupplicated name
    AnoncredsMasterSecretDuplicateNameError = 404,

    AnoncredsProofRejected = 405,

    # Signus errors
    # Unknown format of DID entity keys
    SignusUnknownCryptoError = 500


class IndyError(Exception):
    error_code: ErrorCode

    def __init__(self, error_code: ErrorCode):
        self.error_code = error_code



from .libindy import do_call, create_cb

from typing import Optional
from ctypes import *

import logging


async def sign_and_submit_request(pool_handle: int,
                                  wallet_handle: int,
                                  submitter_did: str,
                                  request_json: str) -> str:
    """
    Signs and submits request message to validator pool.

    Adds submitter information to passed request json, signs it with submitter
    sign key (see wallet_sign), and sends signed request message
    to validator pool (see write_request).

    :param pool_handle: pool handle (created by open_pool_ledger).
    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did: Id of Identity stored in secured Wallet.
    :param request_json: Request data json.
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("sign_and_submit_request: >>> pool_handle: %r, wallet_handle: %r, submitter_did: %r, request_json: %r",
                 pool_handle,
                 wallet_handle,
                 submitter_did,
                 request_json)

    if not hasattr(sign_and_submit_request, "cb"):
        logger.debug("sign_and_submit_request: Creating callback")
        sign_and_submit_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_pool_handle = c_int32(pool_handle)
    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_request_json = c_char_p(request_json.encode('utf-8'))

    request_result = await do_call('indy_sign_and_submit_request',
                                   c_pool_handle,
                                   c_wallet_handle,
                                   c_submitter_did,
                                   c_request_json,
                                   sign_and_submit_request.cb)

    res = request_result.decode()
    logger.debug("sign_and_submit_request: <<< res: %r", res)
    return res


async def submit_request(pool_handle: int,
                         request_json: str) -> str:
    """
    Publishes request message to validator pool (no signing, unlike sign_and_submit_request).
    The request is sent to the validator pool as is. It's assumed that it's already prepared.

    :param pool_handle: pool handle (created by open_pool_ledger).
    :param request_json: Request data json.
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("submit_request: >>> pool_handle: %r, request_json: %r",
                 pool_handle,
                 request_json)

    if not hasattr(submit_request, "cb"):
        logger.debug("submit_request: Creating callback")
        submit_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_pool_handle = c_int32(pool_handle)
    c_request_json = c_char_p(request_json.encode('utf-8'))

    request_result = await do_call('indy_submit_request',
                                   c_pool_handle,
                                   c_request_json,
                                   submit_request.cb)

    res = request_result.decode()
    logger.debug("submit_request: <<< res: %r", res)
    return res


async def sign_request(wallet_handle: int,
                       submitter_did: str,
                       request_json: str) -> str:
    """
    Signs request message.

    Adds submitter information to passed request json, signs it with submitter
    sign key (see wallet_sign).

    :param wallet_handle: wallet handle (created by open_wallet).
    :param submitter_did: Id of Identity stored in secured Wallet.
    :param request_json: Request data json.
    :return: Signed request json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("sign_request: >>> wallet_handle: %r, submitter_did: %r, request_json: %r",
                 wallet_handle,
                 submitter_did,
                 request_json)

    if not hasattr(sign_request, "cb"):
        logger.debug("sign_and_submit_request: Creating callback")
        sign_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_request_json = c_char_p(request_json.encode('utf-8'))

    request_result = await do_call('indy_sign_request',
                                   c_wallet_handle,
                                   c_submitter_did,
                                   c_request_json,
                                   sign_request.cb)

    res = request_result.decode()
    logger.debug("sign_request: <<< res: %r", res)
    return res


async def build_get_ddo_request(submitter_did: str,
                                target_did: str) -> str:
    """
    Builds a request to get a DDO.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_ddo_request: >>> submitter_did: %r, target_did: %r",
                 submitter_did,
                 target_did)

    if not hasattr(build_get_ddo_request, "cb"):
        logger.debug("build_get_ddo_request: Creating callback")
        build_get_ddo_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))

    request_json = await do_call('indy_build_get_ddo_request',
                                 c_submitter_did,
                                 c_target_did,
                                 build_get_ddo_request.cb)

    res = request_json.decode()
    logger.debug("build_get_ddo_request: <<< res: %r", res)
    return res


async def build_nym_request(submitter_did: str,
                            target_did: str,
                            ver_key: Optional[str],
                            alias: Optional[str],
                            role: Optional[str]) -> str:
    """
    Builds a NYM request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :param ver_key: verification key
    :param alias: alias
    :param role: Role of a user NYM record
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_nym_request: >>> submitter_did: %r, target_did: %r, ver_key: %r, alias: %r, role: %r",
                 submitter_did,
                 target_did,
                 ver_key,
                 alias,
                 role)

    if not hasattr(build_nym_request, "cb"):
        logger.debug("build_nym_request: Creating callback")
        build_nym_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))
    c_ver_key = c_char_p(ver_key.encode('utf-8')) if ver_key is not None else None
    c_alias = c_char_p(alias.encode('utf-8')) if alias is not None else None
    c_role = c_char_p(role.encode('utf-8')) if role is not None else None

    request_json = await do_call('indy_build_nym_request',
                                 c_submitter_did,
                                 c_target_did,
                                 c_ver_key,
                                 c_alias,
                                 c_role,
                                 build_nym_request.cb)

    res = request_json.decode()
    logger.debug("build_nym_request: <<< res: %r", res)
    return res


async def build_attrib_request(submitter_did: str,
                               target_did: str,
                               xhash: Optional[str],
                               raw: Optional[str],
                               enc: Optional[str]) -> str:
    """
    Builds an ATTRIB request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :param hash: Hash of attribute data
    :param raw: represented as json, where key is attribute name and value is it's value
    :param enc: Encrypted attribute data
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_attrib_request: >>> submitter_did: %r, target_did: %r, hash: %r, raw: %r, enc: %r",
                 submitter_did,
                 target_did,
                 xhash,
                 raw,
                 enc)

    if not hasattr(build_attrib_request, "cb"):
        logger.debug("build_attrib_request: Creating callback")
        build_attrib_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))
    c_hash = c_char_p(xhash.encode('utf-8')) if xhash is not None else None
    c_raw = c_char_p(raw.encode('utf-8')) if raw is not None else None
    c_enc = c_char_p(enc.encode('utf-8')) if enc is not None else None

    request_json = await do_call('indy_build_attrib_request',
                                 c_submitter_did,
                                 c_target_did,
                                 c_hash,
                                 c_raw,
                                 c_enc,
                                 build_attrib_request.cb)

    res = request_json.decode()
    logger.debug("build_attrib_request: <<< res: %r", res)
    return res


async def build_get_attrib_request(submitter_did: str,
                                   target_did: str,
                                   data: str) -> str:
    """
    Builds a GET_ATTRIB request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :param data: name (attribute name)
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_attrib_request: >>> submitter_did: %r, target_did: %r, data: %r",
                 submitter_did,
                 target_did,
                 data)

    if not hasattr(build_get_attrib_request, "cb"):
        logger.debug("build_get_attrib_request: Creating callback")
        build_get_attrib_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))
    c_data = c_char_p(data.encode('utf-8'))

    request_json = await do_call('indy_build_get_attrib_request',
                                 c_submitter_did,
                                 c_target_did,
                                 c_data,
                                 build_get_attrib_request.cb)

    res = request_json.decode()
    logger.debug("build_get_attrib_request: <<< res: %r", res)
    return res


async def build_get_nym_request(submitter_did: str,
                                target_did: str) -> str:
    """
    Builds a GET_NYM request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_nym_request: >>> submitter_did: %r, target_did: %r",
                 submitter_did,
                 target_did)

    if not hasattr(build_get_nym_request, "cb"):
        logger.debug("build_get_nym_request: Creating callback")
        build_get_nym_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))

    request_json = await do_call('indy_build_get_nym_request',
                                 c_submitter_did,
                                 c_target_did,
                                 build_get_nym_request.cb)

    res = request_json.decode()
    logger.debug("build_get_nym_request: <<< res: %r", res)
    return res


async def build_schema_request(submitter_did: str,
                               data: str) -> str:
    """
    Builds a SCHEMA request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param data: name, version, type, attr_names (ip, port, keys)
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_schema_request: >>> submitter_did: %r, data: %r",
                 submitter_did,
                 data)

    if not hasattr(build_schema_request, "cb"):
        logger.debug("build_schema_request: Creating callback")
        build_schema_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_data = c_char_p(data.encode('utf-8'))

    request_json = await do_call('indy_build_schema_request',
                                 c_submitter_did,
                                 c_data,
                                 build_schema_request.cb)

    res = request_json.decode()
    logger.debug("build_schema_request: <<< res: %r", res)
    return res


async def build_get_schema_request(submitter_did: str,
                                   dest: str,
                                   data: str) -> str:
    """
    Builds a GET_SCHEMA request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param dest: Id of Identity stored in secured Wallet.
    :param data: name, version
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_schema_request: >>> submitter_did: %r, dest: %r, data: %r",
                 submitter_did,
                 dest,
                 data)

    if not hasattr(build_get_schema_request, "cb"):
        logger.debug("build_get_schema_request: Creating callback")
        build_get_schema_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_dest = c_char_p(dest.encode('utf-8'))
    c_data = c_char_p(data.encode('utf-8'))

    request_json = await do_call('indy_build_get_schema_request',
                                 c_submitter_did,
                                 c_dest,
                                 c_data,
                                 build_get_schema_request.cb)

    res = request_json.decode()
    logger.debug("build_get_schema_request: <<< res: %r", res)
    return res


async def build_claim_def_txn(submitter_did: str,
                              xref: int,
                              signature_type: str,
                              data: str) -> str:
    """
    Builds an CLAIM_DEF request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param xref: Seq. number of schema
    :param signature_type: signature type (only CL supported now)
    :param data: components of a key in json: N, R, S, Z
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_schema_request: >>> submitter_did: %r, xref: %r, signature_type: %r, data: %r",
                 submitter_did,
                 xref,
                 signature_type,
                 data)

    if not hasattr(build_claim_def_txn, "cb"):
        logger.debug("build_claim_def_txn: Creating callback")
        build_claim_def_txn.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_xref = c_int32(xref)
    c_signature_type = c_char_p(signature_type.encode('utf-8'))
    c_data = c_char_p(data.encode('utf-8'))

    request_result = await do_call('indy_build_claim_def_txn',
                                   c_submitter_did,
                                   c_xref,
                                   c_signature_type,
                                   c_data,
                                   build_claim_def_txn.cb)

    res = request_result.decode()
    logger.debug("build_claim_def_txn: <<< res: %r", res)
    return res


async def build_get_claim_def_txn(submitter_did: str,
                                  xref: int,
                                  signature_type: str,
                                  origin: str) -> str:
    """
    Builds a GET_CLAIM_DEF request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param xref: Seq. number of schema
    :param signature_type: signature type (only CL supported now)
    :param origin: issuer did
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_claim_def_txn: >>> submitter_did: %r, xref: %r, signature_type: %r, origin: %r",
                 submitter_did,
                 xref,
                 signature_type,
                 origin)

    if not hasattr(build_get_claim_def_txn, "cb"):
        logger.debug("build_get_claim_def_txn: Creating callback")
        build_get_claim_def_txn.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_xref = c_int32(xref)
    c_signature_type = c_char_p(signature_type.encode('utf-8'))
    c_origin = c_char_p(origin.encode('utf-8'))

    request_json = await do_call('indy_build_get_claim_def_txn',
                                 c_submitter_did,
                                 c_xref,
                                 c_signature_type,
                                 c_origin,
                                 build_get_claim_def_txn.cb)

    res = request_json.decode()
    logger.debug("build_get_claim_def_txn: <<< res: %r", res)
    return res


async def build_node_request(submitter_did: str,
                             target_did: str,
                             data: str) -> str:
    """
    Builds a NODE request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param target_did: Id of Identity stored in secured Wallet.
    :param data: id of a target NYM record
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_node_request: >>> submitter_did: %r, target_did: %r, data: %r",
                 submitter_did,
                 target_did,
                 data)

    if not hasattr(build_node_request, "cb"):
        logger.debug("build_node_request: Creating callback")
        build_node_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_target_did = c_char_p(target_did.encode('utf-8'))
    c_data = c_char_p(data.encode('utf-8'))

    request_json = await do_call('indy_build_node_request',
                                 c_submitter_did,
                                 c_target_did,
                                 c_data,
                                 build_node_request.cb)

    res = request_json.decode()
    logger.debug("build_node_request: <<< res: %r", res)
    return res


async def build_get_txn_request(submitter_did: str,
                                data: int) -> str:
    """
    Builds a GET_TXN request.

    :param submitter_did: Id of Identity stored in secured Wallet.
    :param data: seq_no of transaction in ledger
    :return: Request result as json.
    """

    logger = logging.getLogger(__name__)
    logger.debug("build_get_txn_request: >>> submitter_did: %r, data: %r",
                 submitter_did,
                 data)

    if not hasattr(build_get_txn_request, "cb"):
        logger.debug("build_get_txn_request: Creating callback")
        build_get_txn_request.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_submitter_did = c_char_p(submitter_did.encode('utf-8'))
    c_data = c_int32(data)

    request_json = await do_call('indy_build_get_txn_request',
                                 c_submitter_did,
                                 c_data,
                                 build_get_txn_request.cb)

    res = request_json.decode()
    logger.debug("build_get_txn_request: <<< res: %r", res)
    return res



import asyncio
import logging
from ctypes import *
from typing import List, Tuple, Any

from .error import ErrorCode, IndyError
from .libindy import do_call, create_cb


class Event:
    """
    Base class for agent 2 agent communication events

    :handle: (int) Event source handle
    :error: (IndyError) If event is erroneous contains related IndyError exception
    """

    handle: int
    error: IndyError

    def is_success(self):
        """
        Checks is event erroneous or not

        :return: True if there are no errors assigned to this event and False otherwise
        """

        return self.error is None


class ConnectionEvent(Event):
    """
    Agent 2 agent communication listener's connection event

    :connection_handle: Incoming connection handle
    :sender_did: Sender DID
    :receiver_did: Receiver DID
    """
    connection_handle: int
    sender_did: str
    receiver_did: str

    def __init__(self, handle: int, err: int, connection_handle: int, sender_did: bytes, receiver_did: bytes):
        logger = logging.getLogger(__name__)
        logger.debug("ConnectionEvent:__init__ >>> handle: %r, err: %r, connection_handle: %r, sender_did: %r, "
                     "receiver_did: %r",
                     handle,
                     err,
                     connection_handle,
                     sender_did,
                     receiver_did)

        self.handle = handle

        if err != ErrorCode.Success:
            self.error = IndyError(ErrorCode(err))
        else:
            self.connection_handle = connection_handle
            self.sender_did = sender_did.decode()
            self.receiver_did = receiver_did.decode()

        logger.debug("ConnectionEvent:__init__ <<< self: %r", self)


class MessageEvent(Event):
    """
    Agent 2 agent communication connection's message event

    :message: Incoming message
    """

    message: str

    def __init__(self, handle: int, err: int, message: bytes):
        logger = logging.getLogger(__name__)
        logger.debug("MessageEvent:__init__ >>> handle: %r, err: %r, message: %r",
                     handle,
                     err,
                     message)

        self.handle = handle

        if err != ErrorCode.Success:
            self._error = IndyError(ErrorCode(err))
        else:
            self.message = message.decode()

        logger.debug("MessageEvent:__init__ <<< self: %r", self)


_events: List[Event] = []
_event_waiters: List[Tuple[List[int], Any, Any]] = []


def _notify_event_waiters():
    logger = logging.getLogger(__name__)
    logger.debug("_notify_event_waiters: >>> _event_waiters: %r, _events: %r",
                 _event_waiters,
                 _events)

    for i, (handles, event_loop, future) in enumerate(_event_waiters):
        for j, event in enumerate(_events):
            if event.handle in handles:
                del _event_waiters[i]
                del _events[j]
                event_loop.call_soon_threadsafe(lambda f, e: f.set_result(e),
                                                future,
                                                event)
                logger.debug("_notify_event_waiters: <<< handles: %r, event: %r", handles, event)
                return

    logger.debug("_notify_event_waiters: <<< no events")


async def agent_wait_for_event(handles: List[int]) -> Event:
    """
    Waits for events for listeners and connections defined by list of corresponded handles

    :param handles: list of listeners or connections handles
    :return: first occurred listener or connection event
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_wait_for_event: >>> handles: %r", handles)

    event_loop = asyncio.get_event_loop()
    future = event_loop.create_future()

    _event_waiters.append((handles, event_loop, future))
    _notify_event_waiters()

    res = await future

    logger.debug("agent_wait_for_event: <<< res: %r", res)
    return res


async def agent_connect(pool_handle: int,
                        wallet_handle: int,
                        sender_did: str,
                        receiver_did: str) -> int:
    """
    Establishes agent to agent connection.

    Information about sender Identity must be saved in the wallet with indy_create_and_store_my_did
    call before establishing of connection.

    Information about receiver Identity can be saved in the wallet with indy_store_their_did
    call before establishing of connection. If there is no corresponded wallet record for receiver Identity
    than this call will lookup Identity Ledger and cache this information in the wallet.

    Note that messages encryption/decryption will be performed automatically.

    After connection is established returned connection handle can be used to wait for messages with
    agent_wait_for_event or sending messages with agent_send.

    :param pool_handle: pool handle (created by open_pool_ledger).
    :param wallet_handle: wallet handle (created by open_wallet).
    :param sender_did: id of sender Identity stored in secured Wallet.
    :param receiver_did: id of receiver Identity.
    :return: connection handle to use for messages sending and waiting of incoming messages with agent_wait_for_event
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_connect: >>> pool_handle: %r, wallet_handle: %r, sender_did: %r, receiver_did: %r",
                 pool_handle,
                 wallet_handle,
                 sender_did,
                 receiver_did)

    if not hasattr(agent_connect, "connection_cb"):
        logger.debug("agent_connect: Creating connection callback")
        agent_connect.connection_cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_int32))

    def _message_cb(connection_handle: int, err: int, message: bytes):
        logger.debug("agent_connect._message_cb: connection_handle: %r, err: %r, message: %r",
                     connection_handle,
                     err,
                     message)
        _events.append(MessageEvent(connection_handle, err, message))
        _notify_event_waiters()

    if not hasattr(agent_connect, "message_cb"):
        logger.debug("agent_connect: Creating message callback")
        agent_connect.message_cb = CFUNCTYPE(None, c_int32, c_int32, c_char_p)(_message_cb)

    c_pool_handle = c_int32(pool_handle)
    c_wallet_handle = c_int32(wallet_handle)
    c_sender_did = c_char_p(sender_did.encode('utf-8'))
    c_receiver_did = c_char_p(receiver_did.encode('utf-8'))

    res = await do_call('indy_agent_connect',
                        c_pool_handle,
                        c_wallet_handle,
                        c_sender_did,
                        c_receiver_did,
                        agent_connect.connection_cb,
                        agent_connect.message_cb)

    logger.debug("agent_connect: <<< res: %r", res)
    return res


async def agent_listen(endpoint: str) -> int:
    """
    Starts listening of agent connections.

    Listener will accept only connections to registered DIDs by indy_agent_add_identity call.

    Information about sender Identity for incomming connection validation can be saved in the wallet
    with indy_store_their_did call before establishing of connection. If there is no corresponded
    wallet record for sender Identity than listener will lookup Identity Ledger and cache this
    information in the wallet.

    Note that messages encryption/decryption will be performed automatically.

    :param endpoint: endpoint to use in starting listener.
    :return: listener handle to use for waiting of incoming connections with agent_wait_for_event and management of
             assigned to this endpoint identities with agent_add_identity or agent_remove_identity.
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_listen: >>> endpoint: %r", endpoint)

    if not hasattr(agent_listen, "listener_cb"):
        logger.debug("agent_listen: Creating listener callback")
        agent_listen.listener_cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_int32))

    def _connection_cb(listener_handle: int, err: int, connection_handle: int, sender_did: bytes, receiver_did: bytes):
        logger.debug("agent_connect._connection_cb: listener_handle: %r, err: %r, connection_handle: %r, sender_did: "
                     "%r, receiver_did: %r",
                     connection_handle,
                     err,
                     connection_handle,
                     sender_did,
                     receiver_did)
        _events.append(ConnectionEvent(listener_handle, err, connection_handle, sender_did, receiver_did))
        _notify_event_waiters()

    if not hasattr(agent_listen, "connection_cb"):
        logger.debug("agent_listen: Creating connection callback")
        agent_listen.connection_cb = CFUNCTYPE(None, c_int32, c_int32, c_int32, c_char_p, c_char_p)(_connection_cb)

    def _message_cb(connection_handle: int, err: int, message: bytes):
        logger.debug("agent_connect._message_cb: connection_handle: %r, err: %r, message: %r",
                     connection_handle,
                     err,
                     message)
        _events.append(MessageEvent(connection_handle, err, message))
        _notify_event_waiters()

    if not hasattr(agent_listen, "message_cb"):
        logger.debug("agent_connect: Creating message callback")
        agent_listen.message_cb = CFUNCTYPE(None, c_int32, c_int32, c_char_p)(_message_cb)

    c_endpoint = c_char_p(endpoint.encode('utf-8'))

    res = await do_call('indy_agent_listen',
                        c_endpoint,
                        agent_listen.listener_cb,
                        agent_listen.connection_cb,
                        agent_listen.message_cb)

    logger.debug("agent_listen: <<< res: %r", res)
    return res


async def agent_add_identity(listener_handle: int,
                             pool_handle: int,
                             wallet_handle: int,
                             did: str) -> None:
    """
    Add identity to listener.

    Performs wallet lookup to find corresponded receiver Identity information.
    Information about receiver Identity must be saved in the wallet with
    indy_create_and_store_my_did call before this call.

    After successfully add_identity listener will start to accept incoming connection to added DID.

    :param listener_handle: listener handle (created by indy_agent_listen).
    :param pool_handle: pool handle (created by open_pool_ledger).
    :param wallet_handle: wallet handle (created by open_wallet).
    :param did: DID of identity.
    """
    logger = logging.getLogger(__name__)
    logger.debug("agent_add_identity: >>> listener_handle: %r, pool_handle: %r, wallet_handle: %r, did: %r",
                 listener_handle,
                 pool_handle,
                 wallet_handle,
                 did)

    if not hasattr(agent_add_identity, "cb"):
        logger.debug("agent_add_identity: Creating callback")
        agent_add_identity.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_listener_handle = c_int32(listener_handle)
    c_pool_handle = c_int32(pool_handle)
    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    await do_call('indy_agent_add_identity',
                  c_listener_handle,
                  c_pool_handle,
                  c_wallet_handle,
                  c_did,
                  agent_add_identity.cb)

    logger.debug("agent_add_identity: <<<")


async def agent_remove_identity(listener_handle: int,
                                wallet_handle: int,
                                did: str) -> None:
    """
    Remove identity from listener.

    Performs wallet lookup to find corresponded receiver Identity information.
    Information about receiver Identity must be saved in the wallet with
    signus.create_and_store_my_did call before this call.

    After successfully rm_identity listener will stop to accept incoming connection to removed DID.

    :param listener_handle: listener handle (created by indy_agent_listen).
    :param wallet_handle: wallet handle (created by open_wallet).
    :param did: DID of identity.
    """
    logger = logging.getLogger(__name__)
    logger.debug("agent_remove_identity: >>> listener_handle: %r, wallet_handle: %r, did: %r",
                 listener_handle,
                 wallet_handle,
                 did)

    if not hasattr(agent_remove_identity, "cb"):
        logger.debug("agent_remove_identity: Creating callback")
        agent_remove_identity.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_listener_handle = c_int32(listener_handle)
    c_wallet_handle = c_int32(wallet_handle)
    c_did = c_char_p(did.encode('utf-8'))

    await do_call('indy_agent_remove_identity',
                  c_listener_handle,
                  c_wallet_handle,
                  c_did,
                  agent_remove_identity.cb)

    logger.debug("agent_remove_identity: <<<")


async def agent_send(connection_handle: int, message: str) -> None:
    """"
    Sends message to connected agent.

    Note that this call works for both incoming and outgoing connections.
    Note that messages encryption/decryption will be performed automatically.

    :param connection_handle: connection handle returned by indy_agent_connect or indy_agent_listen calls.
    :param message: message to send.
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_send: >>> connection_handle: %r, message: %r",
                 connection_handle,
                 message)

    if not hasattr(agent_send, "cb"):
        logger.debug("agent_send: Creating callback")
        agent_send.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_connection_handle = c_int32(connection_handle)
    c_message = c_char_p(message.encode('utf-8'))

    await do_call('indy_agent_send',
                  c_connection_handle,
                  c_message,
                  agent_send.cb)

    logger.debug("agent_send: <<<")


async def agent_close_connection(connection_handle: int) -> None:
    """
    Closes agent connection.

    Note that this call works for both incoming and outgoing connections.

    :param connection_handle: connection handle returned by indy_agent_connect or indy_agent_listen calls.
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_close_connection: >>> connection_handle: %r", connection_handle)

    if not hasattr(agent_close_connection, "cb"):
        logger.debug("agent_close_connection: Creating callback")
        agent_close_connection.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_connection_handle = c_int32(connection_handle)

    await do_call('indy_agent_close_connection',
                  c_connection_handle,
                  agent_close_connection.cb)

    logger.debug("agent_close_connection: <<<")


async def agent_close_listener(listener_handle: int) -> None:
    """
    Closes listener and stops listening for agent connections.

    Note that all opened incomming connections will be closed automatically.

    :param listener_handle: Listener handle returned by indy_agent_listen call.
    """

    logger = logging.getLogger(__name__)
    logger.debug("agent_close_listener: >>> listener_handle: %r", listener_handle)

    if not hasattr(agent_close_listener, "cb"):
        logger.debug("agent_close_listener: Creating callback")
        agent_close_listener.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_listener_handle = c_int32(listener_handle)

    await do_call('indy_agent_close_listener',
                  c_listener_handle,
                  agent_close_listener.cb)

    logger.debug("agent_close_listener: <<<")



from .libindy import do_call, create_cb

from typing import Optional
from ctypes import *

import logging


async def issuer_create_and_store_claim_def(wallet_handle: int,
                                            issuer_did: str,
                                            schema_json: str,
                                            signature_type: Optional[str],
                                            create_non_revoc: bool) -> str:
    """
    Create keys (both primary and revocation) for the given schema
    and signature type (currently only CL signature type is supported).
    Store the keys together with signature type and schema in a secure wallet as a claim definition.
    The claim definition in the wallet is identifying by a returned unique key.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param issuer_did: a DID of the issuer signing claim_def transaction to the Ledger
    :param schema_json: schema as a json
    :param signature_type: signature type (optional). Currently only 'CL' is supported.
    :param create_non_revoc: whether to request non-revocation claim.
    :return: claim definition json containing information about signature type, schema and issuer's public key.
            Unique number identifying the public key in the wallet
    """

    logger = logging.getLogger(__name__)
    logger.debug("issuer_create_and_store_claim_def: >>> wallet_handle: %r, issuer_did: %r, schema_json: %r,"
                 " signature_type: %r, create_non_revoc: %r",
                 wallet_handle,
                 issuer_did,
                 schema_json,
                 signature_type,
                 create_non_revoc)

    if not hasattr(issuer_create_and_store_claim_def, "cb"):
        logger.debug("issuer_create_and_store_claim_def: Creating callback")
        issuer_create_and_store_claim_def.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_issuer_did = c_char_p(issuer_did.encode('utf-8'))
    c_schema_json = c_char_p(schema_json.encode('utf-8'))
    c_signature_type = c_char_p(signature_type.encode('utf-8')) if signature_type is not None else None
    c_create_non_revoc = c_bool(create_non_revoc)

    claim_def_json = await do_call('indy_issuer_create_and_store_claim_def',
                                   c_wallet_handle,
                                   c_issuer_did,
                                   c_schema_json,
                                   c_signature_type,
                                   c_create_non_revoc,
                                   issuer_create_and_store_claim_def.cb)
    res = claim_def_json.decode()
    logger.debug("issuer_create_and_store_claim_def: <<< res: %r", res)
    return res


async def issuer_create_and_store_revoc_reg(wallet_handle: int,
                                            issuer_did: str,
                                            schema_seq_no: int,
                                            max_claim_num: int) -> str:
    """
    Create a new revocation registry for the given claim definition.
    Stores it in a secure wallet identifying by the returned key.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param issuer_did: a DID of the issuer signing revoc_reg transaction to the Ledger
    :param schema_seq_no: seq no of a schema transaction in Ledger
    :param max_claim_num: maximum number of claims the new registry can process.
    :return: Revoc registry json
    """

    logger = logging.getLogger(__name__)
    logger.debug("issuer_create_and_store_revoc_reg: >>> wallet_handle: %r, issuer_did: %r, schema_seq_no: %r,"
                 " max_claim_num: %r",
                 wallet_handle,
                 issuer_did,
                 schema_seq_no,
                 max_claim_num)

    if not hasattr(issuer_create_and_store_revoc_reg, "cb"):
        logger.debug("issuer_create_and_store_revoc_reg: Creating callback")
        issuer_create_and_store_revoc_reg.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_issuer_did = c_char_p(issuer_did.encode('utf-8'))
    c_schema_seq_no = c_int32(schema_seq_no)
    c_max_claim_num = c_int32(max_claim_num)

    revoc_reg_json = await do_call('indy_issuer_create_and_store_revoc_reg',
                                   c_wallet_handle,
                                   c_issuer_did,
                                   c_schema_seq_no,
                                   c_max_claim_num,
                                   issuer_create_and_store_revoc_reg.cb)
    res = revoc_reg_json.decode()
    logger.debug("issuer_create_and_store_revoc_reg: <<< res: %r", res)
    return res


async def issuer_create_claim(wallet_handle: int,
                              claim_req_json: str,
                              claim_json: str,
                              user_revoc_index: int) -> (str, str):
    """
    Signs a given claim for the given user by a given key (claim ef).
    The corresponding claim definition and revocation registry must be already created
    an stored into the wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param claim_req_json: a claim request with a blinded secret
        from the user (returned by prover_create_and_store_claim_req).
        Also contains schema_seq_no and issuer_did
        Example:
        {
            "blinded_ms" : <blinded_master_secret>,
            "schema_seq_no" : <schema_seq_no>,
            "issuer_did" : <issuer_did>
        }
    :param claim_json: a claim containing attribute values for each of requested attribute names.
        Example:
        {
            "attr1" : ["value1", "value1_as_int"],
            "attr2" : ["value2", "value2_as_int"]
        }
    :param user_revoc_index: index of a new user in the revocation registry
     (optional, pass -1 if user_revoc_index is absentee; default one is used if not provided)
    :return: Revocation registry update json with a newly issued claim
        Claim json containing issued claim, issuer_did, schema_seq_no, and revoc_reg_seq_no
        used for issuance
        {
            "claim": <see claim_json above>,
            "signature": <signature>,
            "revoc_reg_seq_no", string,
            "issuer_did", string,
            "schema_seq_no", string,
        }
    """

    logger = logging.getLogger(__name__)
    logger.debug("issuer_create_claim: >>> wallet_handle: %r, claim_req_json: %r, claim_json: %r,"
                 " revoc_reg_seq_no: %r, user_revoc_index: %r",
                 wallet_handle,
                 claim_req_json,
                 claim_json,
                 user_revoc_index)

    if not hasattr(issuer_create_claim, "cb"):
        logger.debug("issuer_create_claim: Creating callback")
        issuer_create_claim.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_claim_req_json = c_char_p(claim_req_json.encode('utf-8'))
    c_claim_json = c_char_p(claim_json.encode('utf-8'))
    c_user_revoc_index = c_int32(user_revoc_index)

    (revoc_reg_update_json, claim_json) = await do_call('indy_issuer_create_claim',
                                                        c_wallet_handle,
                                                        c_claim_req_json,
                                                        c_claim_json,
                                                        c_user_revoc_index,
                                                        issuer_create_claim.cb)
    res = (revoc_reg_update_json.decode(), claim_json.decode())
    logger.debug("issuer_create_claim: <<< res: %r", res)
    return res


async def issuer_revoke_claim(wallet_handle: int,
                              issuer_did: str,
                              schema_seq_no: int,
                              user_revoc_index: int) -> str:
    """
    Revokes a user identified by a revoc_id in a given revoc-registry.
    The corresponding claim definition and revocation registry must be already
    created an stored into the wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param issuer_did: a DID of the issuer signing claim_def transaction to the Ledger
    :param schema_seq_no: seq no of a schema transaction in Ledger
    :param user_revoc_index: index of the user in the revocation registry
    :return: Revocation registry update json with a revoked claim
    """

    logger = logging.getLogger(__name__)
    logger.debug("issuer_revoke_claim: >>> wallet_handle: %r, revoc_reg_seq_no: %r, user_revoc_index: %r",
                 wallet_handle,
                 issuer_did,
                 schema_seq_no,
                 user_revoc_index)

    if not hasattr(issuer_revoke_claim, "cb"):
        logger.debug("issuer_revoke_claim: Creating callback")
        issuer_revoke_claim.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_issuer_did = c_char_p(issuer_did.encode('utf-8'))
    c_schema_seq_no = c_int32(schema_seq_no)
    c_user_revoc_index = c_int32(user_revoc_index)

    revoc_reg_update_json = await do_call('indy_issuer_revoke_claim',
                                          c_wallet_handle,
                                          c_issuer_did,
                                          c_schema_seq_no,
                                          c_user_revoc_index,
                                          issuer_revoke_claim.cb)
    res = revoc_reg_update_json.decode()
    logger.debug("issuer_revoke_claim: <<< res: %r", res)
    return res


async def prover_store_claim_offer(wallet_handle: int,
                                   claim_offer_json: str) -> None:
    """
    Stores a claim offer from the given issuer in a secure storage.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param claim_offer_json: claim offer as a json containing information about the issuer and a claim:
        {
            "issuer_did": string,
            "schema_seq_no": string
        }
    :return: None.
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_store_claim_offer: >>> wallet_handle: %r, claim_offer_json: %r",
                 wallet_handle,
                 claim_offer_json)

    if not hasattr(prover_store_claim_offer, "cb"):
        logger.debug("prover_store_claim_offer: Creating callback")
        prover_store_claim_offer.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_claim_offer_json = c_char_p(claim_offer_json.encode('utf-8'))

    res = await do_call('indy_prover_store_claim_offer',
                        c_wallet_handle,
                        c_claim_offer_json,
                        prover_store_claim_offer.cb)

    logger.debug("prover_store_claim_offer: <<< res: %r", res)
    return res


async def prover_get_claim_offers(wallet_handle: int,
                                  filter_json: str) -> str:
    """
    Gets all stored claim offers (see prover_store_claim_offer).
    A filter can be specified to get claim offers for specific Issuer, claim_def or schema only.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param filter_json: optional filter to get claim offers for specific Issuer, claim_def or schema only only
        Each of the filters is optional and can be combines
            {
                "issuer_did": string,
                "schema_seq_no": string
            }
    :return: A json with a list of claim offers for the filter.
        {
            [{"issuer_did": string,
            "schema_seq_no": string}]
        }
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_store_claim_offer: >>> wallet_handle: %r, filter_json: %r",
                 wallet_handle,
                 filter_json)

    if not hasattr(prover_get_claim_offers, "cb"):
        logger.debug("prover_get_claim_offers: Creating callback")
        prover_get_claim_offers.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_filter_json = c_char_p(filter_json.encode('utf-8'))

    claim_offers_json = await do_call('indy_prover_get_claim_offers',
                                      c_wallet_handle,
                                      c_filter_json,
                                      prover_get_claim_offers.cb)

    res = claim_offers_json.decode()
    logger.debug("prover_get_claim_offers: <<< res: %r", res)
    return res


async def prover_create_master_secret(wallet_handle: int,
                                      master_secret_name: str) -> None:
    """
    Creates a master secret with a given name and stores it in the wallet.
    The name must be unique.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param master_secret_name: a new master secret name
    :return: None.
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_create_master_secret: >>> wallet_handle: %r, master_secret_name: %r",
                 wallet_handle,
                 master_secret_name)

    if not hasattr(prover_create_master_secret, "cb"):
        logger.debug("prover_create_master_secret: Creating callback")
        prover_create_master_secret.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_master_secret_name = c_char_p(master_secret_name.encode('utf-8'))

    res = await do_call('indy_prover_create_master_secret',
                        c_wallet_handle,
                        c_master_secret_name,
                        prover_create_master_secret.cb)

    logger.debug("prover_create_master_secret: <<< res: %r", res)
    return res


async def prover_create_and_store_claim_req(wallet_handle: int,
                                            prover_did: str,
                                            claim_offer_json: str,
                                            claim_def_json: str,
                                            master_secret_name: str) -> str:
    """
    Creates a clam request json for the given claim offer and stores it in a secure wallet.
    The claim offer contains the information about Issuer (DID, schema_seq_no),
    and the schema (schema_seq_no).
    The method gets public key and schema from the ledger, stores them in a wallet,
    and creates a blinded master secret for a master secret identified by a provided name.
    The master secret identified by the name must be already stored in the secure wallet (see prover_create_master_secret)
    The blinded master secret is a part of the claim request.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param prover_did: a DID of the prover
    :param claim_offer_json: claim offer as a json containing information about the issuer and a claim:
        {
            "issuer_did": string,
            "schema_seq_no": string
        }
    :param claim_def_json: claim definition json associated with issuer_did and schema_seq_no in the claim_offer
    :param master_secret_name: the name of the master secret stored in the wallet
    :return: Claim request json.
        {
            "blinded_ms" : <blinded_master_secret>,
            "schema_seq_no" : <schema_seq_no>,
            "issuer_did" : <issuer_did>
        }
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_create_and_store_claim_req: >>> wallet_handle: %r, prover_did: %r, claim_offer_json: %r,"
                 " claim_def_json: %r, master_secret_name: %r",
                 wallet_handle,
                 prover_did,
                 claim_offer_json,
                 claim_def_json,
                 master_secret_name)

    if not hasattr(prover_create_and_store_claim_req, "cb"):
        logger.debug("prover_create_and_store_claim_req: Creating callback")
        prover_create_and_store_claim_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_prover_did = c_char_p(prover_did.encode('utf-8'))
    c_claim_offer_json = c_char_p(claim_offer_json.encode('utf-8'))
    c_claim_def_json = c_char_p(claim_def_json.encode('utf-8'))
    c_master_secret_name = c_char_p(master_secret_name.encode('utf-8'))

    claim_req_json = await do_call('indy_prover_create_and_store_claim_req',
                                   c_wallet_handle,
                                   c_prover_did,
                                   c_claim_offer_json,
                                   c_claim_def_json,
                                   c_master_secret_name,
                                   prover_create_and_store_claim_req.cb)

    res = claim_req_json.decode()
    logger.debug("prover_create_and_store_claim_req: <<< res: %r", res)
    return res


async def prover_store_claim(wallet_handle: int,
                             claims_json: str) -> None:
    """
    Updates the claim by a master secret and stores in a secure wallet.
    The claim contains the information about
    schema_seq_no, issuer_did, revoc_reg_seq_no (see issuer_create_claim).
    Seq_no is a sequence number of the corresponding transaction in the ledger.
    The method loads a blinded secret for this key from the wallet,
    updates the claim and stores it in a wallet.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param claims_json: claim json:
        {
            "claim": {attr1:[value, value_as_int]}
            "signature": <signature>,
            "schema_seq_no": string,
            "revoc_reg_seq_no", string
            "issuer_did", string
        }
    :return: None.
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_store_claim: >>> wallet_handle: %r, claims_json: %r",
                 wallet_handle,
                 claims_json)

    if not hasattr(prover_store_claim, "cb"):
        logger.debug("prover_store_claim: Creating callback")
        prover_store_claim.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32))

    c_wallet_handle = c_int32(wallet_handle)
    c_claims_json = c_char_p(claims_json.encode('utf-8'))

    res = await do_call('indy_prover_store_claim',
                        c_wallet_handle,
                        c_claims_json,
                        prover_store_claim.cb)

    logger.debug("prover_store_claim: <<< res: %r", res)
    return res


async def prover_get_claims(wallet_handle: int,
                            filter_json: str) -> str:
    """
    Gets human readable claims according to the filter.
    If filter is NULL, then all claims are returned.
    Claims can be filtered by Issuer, claim_def and/or Schema.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param filter_json: filter for claims
        {
            "issuer_did": string,
            "schema_seq_no": string
        }
    :return: claims json
        [{
            "claim_uuid": <string>,
            "attrs": [{"attr_name" : "attr_value"}],
            "schema_seq_no": string,
            "issuer_did": string,
            "revoc_reg_seq_no": string,
        }]
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_get_claims: >>> wallet_handle: %r, filter_json: %r",
                 wallet_handle,
                 filter_json)

    if not hasattr(prover_get_claims, "cb"):
        logger.debug("prover_get_claims: Creating callback")
        prover_get_claims.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_filter_json = c_char_p(filter_json.encode('utf-8'))

    claims_json = await do_call('indy_prover_get_claims',
                                c_wallet_handle,
                                c_filter_json,
                                prover_get_claims.cb)

    res = claims_json.decode()
    logger.debug("prover_get_claims: <<< res: %r", res)
    return res


async def prover_get_claims_for_proof_req(wallet_handle: int,
                                          proof_request_json: str) -> str:
    """
    Gets human readable claims matching the given proof request.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param proof_request_json: proof request json
        {
            "name": string,
            "version": string,
            "nonce": string,
            "requested_attr1_uuid": <attr_info>,
            "requested_attr2_uuid": <attr_info>,
            "requested_attr3_uuid": <attr_info>,
            "requested_predicate_1_uuid": <predicate_info>,
            "requested_predicate_2_uuid": <predicate_info>,
        }
    :return: json with claims for the given pool request.
        Claim consists of uuid, human-readable attributes (key-value map), schema_seq_no, issuer_did and revoc_reg_seq_no.
            {
                "requested_attr1_uuid": [claim1, claim2],
                "requested_attr2_uuid": [],
                "requested_attr3_uuid": [claim3],
                "requested_predicate_1_uuid": [claim1, claim3],
                "requested_predicate_2_uuid": [claim2],
            }, where claim is
            {
                "claim_uuid": <string>,
                "attrs": [{"attr_name" : "attr_value"}],
                "schema_seq_no": string,
                "issuer_did": string,
                "revoc_reg_seq_no": string,
            }
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_get_claims_for_proof_req: >>> wallet_handle: %r, proof_request_json: %r",
                 wallet_handle,
                 proof_request_json)

    if not hasattr(prover_get_claims, "cb"):
        logger.debug("prover_get_claims_for_proof_req: Creating callback")
        prover_get_claims_for_proof_req.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_proof_request_json = c_char_p(proof_request_json.encode('utf-8'))

    claims_json = await do_call('indy_prover_get_claims_for_proof_req',
                                c_wallet_handle,
                                c_proof_request_json,
                                prover_get_claims_for_proof_req.cb)

    res = claims_json.decode()
    logger.debug("prover_get_claims_for_proof_req: <<< res: %r", res)
    return res


async def prover_create_proof(wallet_handle: int,
                              proof_req_json: str,
                              requested_claims_json: str,
                              schemas_json: str,
                              master_secret_name: str,
                              claim_defs_json: str,
                              revoc_regs_json: str) -> str:
    """
    Creates a proof according to the given proof request
    Either a corresponding claim with optionally revealed attributes or self-attested attribute must be provided
    for each requested attribute (see indy_prover_get_claims_for_pool_req).
    A proof request may request multiple claims from different schemas and different issuers.
    All required schemas, public keys and revocation registries must be provided.
    The proof request also contains nonce.
    The proof contains either proof or self-attested attribute value for each requested attribute.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param proof_req_json: proof request json as come from the verifier
        {
            "nonce": string,
            "requested_attr1_uuid": <attr_info>,
            "requested_attr2_uuid": <attr_info>,
            "requested_attr3_uuid": <attr_info>,
            "requested_predicate_1_uuid": <predicate_info>,
            "requested_predicate_2_uuid": <predicate_info>,
        }
    :param requested_claims_json: either a claim or self-attested attribute for each requested attribute
        {
            "requested_attr1_uuid": [claim1_uuid_in_wallet, true <reveal_attr>],
            "requested_attr2_uuid": [self_attested_attribute],
            "requested_attr3_uuid": [claim2_seq_no_in_wallet, false]
            "requested_attr4_uuid": [claim2_seq_no_in_wallet, true]
            "requested_predicate_1_uuid": [claim2_seq_no_in_wallet],
            "requested_predicate_2_uuid": [claim3_seq_no_in_wallet],
        }
    :param schemas_json: all schema jsons participating in the proof request
        {
            "claim1_uuid_in_wallet": <schema1>,
            "claim2_uuid_in_wallet": <schema2>,
            "claim3_uuid_in_wallet": <schema3>,
        }
    :param master_secret_name: the name of the master secret stored in the wallet

    :param claim_defs_json: all claim definition jsons participating in the proof request
        {
            "claim1_uuid_in_wallet": <claim_def1>,
            "claim2_uuid_in_wallet": <claim_def2>,
            "claim3_uuid_in_wallet": <claim_def3>,
        }
    :param revoc_regs_json: all revocation registry jsons participating in the proof request
        {
            "claim1_uuid_in_wallet": <revoc_reg1>,
            "claim2_uuid_in_wallet": <revoc_reg2>,
            "claim3_uuid_in_wallet": <revoc_reg3>,
        }
    :return: Proof json
        For each requested attribute either a proof (with optionally revealed attribute value) or
        self-attested attribute value is provided.
        Each proof is associated with a claim and corresponding schema_seq_no, issuer_did and revoc_reg_seq_no.
        There ais also aggregated proof part common for all claim proofs.
            {
                "requested": {
                    "requested_attr1_id": [claim_proof1_uuid, revealed_attr1, revealed_attr1_as_int],
                    "requested_attr2_id": [self_attested_attribute],
                    "requested_attr3_id": [claim_proof2_uuid]
                    "requested_attr4_id": [claim_proof2_uuid, revealed_attr4, revealed_attr4_as_int],
                    "requested_predicate_1_uuid": [claim_proof2_uuid],
                    "requested_predicate_2_uuid": [claim_proof3_uuid],
                }
                "claim_proofs": {
                    "claim_proof1_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no],
                    "claim_proof2_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no],
                    "claim_proof3_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no]
                },
                "aggregated_proof": <aggregated_proof>
            }
    """

    logger = logging.getLogger(__name__)
    logger.debug("prover_create_proof: >>> wallet_handle: %r, proof_req_json: %r,"
                 " requested_claims_json: %r, schemas_json: %r, master_secret_name: %r,"
                 " claim_defs_json: %r, revoc_regs_json: %r",
                 wallet_handle,
                 proof_req_json,
                 requested_claims_json,
                 schemas_json,
                 master_secret_name,
                 claim_defs_json,
                 revoc_regs_json)

    if not hasattr(prover_create_proof, "cb"):
        logger.debug("prover_create_proof: Creating callback")
        prover_create_proof.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_char_p))

    c_wallet_handle = c_int32(wallet_handle)
    c_proof_req_json = c_char_p(proof_req_json.encode('utf-8'))
    c_requested_claims_json = c_char_p(requested_claims_json.encode('utf-8'))
    c_schemas_json = c_char_p(schemas_json.encode('utf-8'))
    c_master_secret_name = c_char_p(master_secret_name.encode('utf-8'))
    c_claim_defs_json = c_char_p(claim_defs_json.encode('utf-8'))
    c_revoc_regs_json = c_char_p(revoc_regs_json.encode('utf-8'))

    proof_json = await do_call('indy_prover_create_proof',
                               c_wallet_handle,
                               c_proof_req_json,
                               c_requested_claims_json,
                               c_schemas_json,
                               c_master_secret_name,
                               c_claim_defs_json,
                               c_revoc_regs_json,
                               prover_create_proof.cb)

    res = proof_json.decode()
    logger.debug("prover_create_proof: <<< res: %r", res)
    return res


async def verifier_verify_proof(proof_request_json: str,
                                proof_json: str,
                                schemas_json: str,
                                claim_defs_jsons: str,
                                revoc_regs_json: str) -> bool:
    """
    Verifies a proof (of multiple claim).
    All required schemas, public keys and revocation registries must be provided.

    :param wallet_handle: wallet handler (created by open_wallet).
    :param proof_request_json: initial proof request as sent by the verifier
        {
            "nonce": string,
            "requested_attr1_uuid": <attr_info>,
            "requested_attr2_uuid": <attr_info>,
            "requested_attr3_uuid": <attr_info>,
            "requested_predicate_1_uuid": <predicate_info>,
            "requested_predicate_2_uuid": <predicate_info>,
        }
    :param proof_json: proof json
        For each requested attribute either a proof (with optionally revealed attribute value) or
        self-attested attribute value is provided.
        Each proof is associated with a claim and corresponding schema_seq_no, issuer_did and revoc_reg_seq_no.
        There ais also aggregated proof part common for all claim proofs.
            {
                "requested": {
                    "requested_attr1_id": [claim_proof1_uuid, revealed_attr1, revealed_attr1_as_int],
                    "requested_attr2_id": [self_attested_attribute],
                    "requested_attr3_id": [claim_proof2_uuid]
                    "requested_attr4_id": [claim_proof2_uuid, revealed_attr4, revealed_attr4_as_int],
                    "requested_predicate_1_uuid": [claim_proof2_uuid],
                    "requested_predicate_2_uuid": [claim_proof3_uuid],
                }
                "claim_proofs": {
                    "claim_proof1_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no],
                    "claim_proof2_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no],
                    "claim_proof3_uuid": [<claim_proof>, issuer_did, schema_seq_no, revoc_reg_seq_no]
                },
                "aggregated_proof": <aggregated_proof>
            }
    :param schemas_json: all schema jsons participating in the proof
        {
            "claim_proof1_uuid": <schema>,
            "claim_proof2_uuid": <schema>,
            "claim_proof3_uuid": <schema>
        }
    :param claim_defs_jsons: all claim definition jsons participating in the proof
        {
            "claim_proof1_uuid": <claim_def>,
            "claim_proof2_uuid": <claim_def>,
            "claim_proof3_uuid": <claim_def>
        }
    :param revoc_regs_json: all revocation registry jsons participating in the proof
        {
            "claim_proof1_uuid": <revoc_reg>,
            "claim_proof2_uuid": <revoc_reg>,
            "claim_proof3_uuid": <revoc_reg>
        }
    :return: valid: true - if signature is valid, false - otherwise
    """

    logger = logging.getLogger(__name__)
    logger.debug("verifier_verify_proof: >>> proof_request_json: %r,"
                 " proof_json: %r, schemas_json: %r, claim_defs_jsons: %r, revoc_regs_json: %r",
                 proof_request_json,
                 proof_json,
                 schemas_json,
                 claim_defs_jsons,
                 revoc_regs_json)

    if not hasattr(verifier_verify_proof, "cb"):
        logger.debug("verifier_verify_proof: Creating callback")
        verifier_verify_proof.cb = create_cb(CFUNCTYPE(None, c_int32, c_int32, c_bool))

    c_proof_request_json = c_char_p(proof_request_json.encode('utf-8'))
    c_proof_json = c_char_p(proof_json.encode('utf-8'))
    c_schemas_json = c_char_p(schemas_json.encode('utf-8'))
    c_claim_defs_jsons = c_char_p(claim_defs_jsons.encode('utf-8'))
    c_revoc_regs_json = c_char_p(revoc_regs_json.encode('utf-8'))

    res = await do_call('indy_verifier_verify_proof',
                        c_proof_request_json,
                        c_proof_json,
                        c_schemas_json,
                        c_claim_defs_jsons,
                        c_revoc_regs_json,
                        verifier_verify_proof.cb)

    logger.debug("verifier_verify_proof: <<< res: %r", res)
    return res



from distutils.core import setup

setup(
    name='indy_sdk',
    version='0.0.1-dev-162',
    packages=['indy'],
    url='https://github.com/hyperledger/indy-sdk',
    license='MIT/Apache-2.0',
    author='Vyacheslav Gudkov',
    author_email='vyacheslav.gudkov@dsr-company.com',
    description='This is the official SDK for Hyperledger Indy (https://www.hyperledger.org/projects), which provides a distributed-ledger-based foundation for self-sovereign identity (https://sovrin.org). The major artifact of the SDK is a c-callable library.',
    install_requires=['pytest', 'pytest-asyncio', 'base58'],
    tests_require=['pytest', 'pytest-asyncio', 'base58']
)
